//----------------------------------------------------------------------//
//  PDP: calculate primary deviation pair of fundamental matrix         //
//                                                                      //
//  2000/2/29                                                           //
//  Programmed by Hitoshi Mishima                                       //
//  Department of Computer Science,                                     //
//  Gunma University Faculty of Engineering                             //
//----------------------------------------------------------------------//

#include <statcomp.h>
#include <tool.h>
#include "correct.h"
#include "extra.h"

int
PDP(int point_num, VectorN eigen_vec, Matrix* eigen_mat, 
    double e_hat2, Matrix& F, Matrix& F_plus, Matrix& F_minus, 
    double eps_d, int crr_flag)
//----------------------------------------------------------------------//
//ǽ                                                                  //
//  ⡼ȥƥ󥽥9Ĥθͭͤбñ̸ͭľ,  //
//  ڤΥ٥οͤͿ, Ŭɸа      //
//  (primary deviation pair) F_plus, F_minus֤.			//
//                                                                      //
//                                                                  //
//  point_num : ǡ                              input		//
//  eigen_vec : ⡼ȥƥ󥽥θͭ              input		//
//  eigen_mat : eigen_valбͭ             input		//
//  e_hat2    : Υ٥ο                input		//
//  F         : Ѥߤδù                      input		//
//  F_plus    : Fɸа(+)                          output		//
//  F_minus   : F      (-)                          output		//
//  eps_d     : ȿλ                            input		//
//              (ǥե: 1.0e-12)					//
//  crr_flag  : Ŭ¹ԥե饰                      input		//
//              (=1ʤFκŬ¹, ǥե: 1)		//
//----------------------------------------------------------------------//
{
   Tensor3333	M_hat, V_0F, W, Pi;
   VectorN	eigen_mu(9);
   Matrix	eigen_G[9], F_trs;
   double	mu;
   int		itr;

   /******⡼ȥƥ󥽥M_hat׻ľ******/
   M_hat.clear(0.0);
   for (int a = 0; a < 8; a++)
      M_hat = M_hat + eigen_vec[a]*TensorProduct(eigen_mat[a], eigen_mat[a]);

   /******ƥ󥽥W׻******/
   W.clear(0.0);
   Pi = Pseudo(Cofactor(F));
   for (int i = 0; i < 3; i++)
      for (int j = 0; j < 3; j++)
	 for (int k = 0; k < 3; k++)
	    for (int l = 0; l < 3; l++)
	       for (int m = 0; m < 3; m++)
		  for (int n = 0; n < 3; n++)
		     for (int p = 0; p < 3; p++)
			for (int q = 0; q < 3; q++)
			   W(i,j,k,l) +=
			      Pi(i,j,m,n) * Pi(k,l,p,q) * M_hat(m,n,p,q);
  

   /******ƥ󥽥W9Ĥθͭͭ͡******/
   if (W.Eigen(eigen_mu, eigen_G))
   {
      //ɸаF_plus, F_minus׻
      mu = sqrt(e_hat2 / (eigen_mu[6] * (double)point_num));
      F_plus  = Normalize(F + mu * eigen_G[6]);
      F_minus = Normalize(F - mu * eigen_G[6]);

      //F_plus, F_minusι󼰤0Ȥʤ褦
      if (crr_flag == 1)
      {
	 // F_plus 
	 itr = 0;
	 while ((fabs(Det(F_plus)) > eps_d) && (itr++ < CRR_MAX_ITR))
	 {
	    F_trs = Trans(Cofactor(F_plus));
	    F_plus = Normalize(F_plus - Det(F_plus)*F_trs/(F_trs, F_trs));
	 }
	 if (itr > CRR_MAX_ITR)
	 {
	    fprintf(stderr,
		    "Warning: the iteration of PDP correction exceeded %d times.\n",
		    CRR_MAX_ITR);
	 }

	 // F_minus 
	 itr = 0;
	 while ((fabs(Det(F_minus)) > eps_d) && (itr++ < CRR_MAX_ITR))
	 {
	    F_trs = Trans(Cofactor(F_minus));
	    F_minus = Normalize(F_minus - Det(F_minus)*F_trs/(F_trs, F_trs));
	 }
	 if (itr > CRR_MAX_ITR)
	 {
	    fprintf(stderr,
		    "Warinng: the iteration of PDP correction exceeded %d times.\n",
		    CRR_MAX_ITR);
	 }
      }
   }
   else
   {
      fprintf(stderr, "Cannot calculate primary deviation pair of F.\n");
      return 0;
   }

   return 1;
}
