//----------------------------------------------------------------------//
//  fundamental: calculate fundamental matrix by renormalization        //
//                                                                      //
//  2000/2/29                                                           //
//  Programmed by Hitoshi Mishima                                       //
//  Department of Computer Science,                                     //
//  Gunma University Faculty of Engineering                             //
//----------------------------------------------------------------------//
//  2002/12/10								//
//  Modified by Yasushi Kanazawa					//
//----------------------------------------------------------------------//

#include <statcomp.h>
#include <tool.h>

//
// ƥƥ󥽥η׻
//
void
set_tensors(Tensor3333& M, Tensor3333& N1, Tensor3333& N2,
	    Vector& x1, Vector& x2, Matrix& V1, Matrix& V2,
	    double W, int dim)
{
   for (int i = 0; i < 3; i++)
      for (int j = 0; j < 3; j++)
	 for (int k = 0; k < 3; k++)
	    for (int l = 0; l < 3; l++)
	    {
	       if (3*(i-1)+j <= 3*(k-1)+l)
	       {
		  M(i,j,k,l) += W*x1[i]*x2[j]*x1[k]*x2[l];
		  N1(i,j, k,l) +=
		     W*(V1[i][k]*x2[j]*x2[l] + V2[j][l]*x1[i]*x1[k]);
		  if (dim == 2)
		     N2(i,j,k,l) += W*V1[i][k]*V2[j][l];
	       }
	    }
}

// ꤳˡ
int Renormal(ImagePosition* P1, ImagePosition* P2, int point_num, 
	     VectorN& eigen_val, Matrix* eigen_mat, double& J, 
	     int dim, double min, int itmax)
//----------------------------------------------------------------------//
//ǽ									//
//  ֤ȤȥǡͿ, ùF򤯤ꤳˡǷ׻.	//
//									//
//									//
//  P1        : 1ǡ                       input		//
//  P2        : 2ǡ                       input		//
//  point_num : ǡ                              input		//
//  eigen_val : ⡼ȥƥ󥽥θͭ              output		//
//  eigen_mat : eigen_valбͭ             output		//
//  J         : ꤳˡλĺ                        output		//
//  dim       : ꤳˡμ                        input		//
//               (1 ޤ 2, ǥե: 1)				//
//  min       : ꤳˡȿλ                input		//
//               (ǥե: 1.0e-8)					//
//  itmax     : ꤳˡκȿ                input		//
//               (>=1, ǥե: 50)				//
//									//
//֤								//
//  ꤳˡȿ						//
//----------------------------------------------------------------------//
{
   int		a, iter = 0;
   double	lambda9, W = 1.0, c = 0.0, J2 = 0.0;
   double	tmpw, inp1, inp2;
   Tensor3333	M_hat, M, N1, N2;
   Vector	x1, x2;
   Matrix	V1, V2, F9, TF9;

   /******ƥ󥽥ȻĺJ******/
   M.clear(0.0);
   N1.clear(0.0);
   if (dim == 2) N2.clear(0.0);
   J = 1.0e+10;

   /******ƥƥ󥽥Ǥ׻******/
   for (a = 0; a < point_num; a++)
   {    
      //֤٥ȥȶʬ
      x1 = P1[a].Pos();      V1 = P1[a].Cov();
      x2 = P2[a].Pos();      V2 = P2[a].Cov();      
      
      //ƥƥ󥽥Ƿ׻
      set_tensors(M, N1, N2, x1, x2, V1, V2, W, dim);
   }


   /******ƥƥ󥽥λĤǤ򤹤٤******/
   M.Fill();
   M = M / (double)point_num;
   N1.Fill();
   N1 = N1 / (double)point_num;
   if (dim == 2)
   {
      N2.Fill();
      N2 = N2 / (double)point_num;
   }
   M_hat = M;

  
   /******M_hat9Ĥθͭͭ͡******/
   if (M_hat.Eigen(eigen_val, eigen_mat) == 0)
      return 0;
   else
   {
      lambda9 = eigen_val[8]; //Ǿͭ
      if (lambda9 < 0.0) lambda9 *= -1.0;
      F9 = eigen_mat[8];      //бΥ1θͭ
   }

   /******ꤳˡȿ******/
   do 
   {
      iter++; //ȿ+1

      //ȿitmaxۤ齪λ(ꤳˡ)
      if (iter > itmax)
	 return 0;

      //cι
      inp1 = (F9, N1*F9); //F9N1*F9
      inp2 = (F9, N2*F9); //F9N2*F9

      if (dim == 2) 
      {
	 double D = pow(inp1 - 2.0*c*inp2, 2) - 4.0*lambda9*inp2;
	 if (D >= 0.0)
	    c += (inp1 - 2.0*c*inp2 - sqrt(D))/(2.0*inp2);
	 else
	    c += lambda9/inp1;
      }
      else 
	 c += lambda9/inp1;

      //ƥƥ󥽥ν
      M.clear(0.0);
      N1.clear(0.0);
      if (dim == 2) 
	 N2.clear(0.0);

      //ƥƥ󥽥Ǥ׻
      for (a = 0; a < point_num; a++)
      {
	 //֤٥ȥȶʬ
	 x1 = P1[a].Pos();      V1 = P1[a].Cov();
	 x2 = P2[a].Pos();      V2 = P2[a].Cov();      
      
	 //ŤWι
	 TF9 = Trans(F9);
	 tmpw = (x2, (TF9*V1*F9*x2)) + (x1, (F9*V2*TF9*x1));
	 if (dim == 2) tmpw += c*(V1*F9, F9*V2);
	 W = 1.0/tmpw;
	  
	 //ƥƥ󥽥Ƿ׻
	 set_tensors(M, N1, N2, x1, x2, V1, V2, W, dim);
      }
      
      //ƥƥ󥽥λĤǤ򤹤٤
      M.Fill();
      N1.Fill();
      if (dim == 2) N2.Fill();

      //ƥƥ󥽥ι
      M = M/(double)(point_num);
      N1 = N1/(double)(point_num);
      M_hat = M - c*N1;
      if (dim == 2)
      {
	 N2 = N2/(double)(point_num);
	 M_hat = M_hat + c*c*N2;
      }

      //ĺJ򹹿
      J2 = J;
      J = (F9, M*F9);
      
      if (J2 < J) J = J2;
      else
      {
	 //M_hat9Ĥθͭͭ͡
	 if (M_hat.Eigen(eigen_val, eigen_mat) == 0)
	    return 0;
	 else
	 {
	    lambda9 = eigen_val[8]; //Ǿͭ
	    F9 = eigen_mat[8];      //бΥ1θͭ
	 }
      }
   } while ((J2 > J) && ((fabs(lambda9)) > min));
  
   return iter;
}
